/*
 * Copyright (c) 2018, Larry Lile <lile@users.sourceforge.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */

#include "config.h"

#include <stdio.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_STRING_H
# if !defined STDC_HEADERS && defined HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_PWD_H
# include <pwd.h>
#endif

#ifdef HAVE_STDARG_H
# include <stdarg.h>
#endif

#ifdef HAVE_GETOPT_H
# include <getopt.h>
#endif

#ifdef HAVE_LDAP_H
# include <ldap.h>
#endif

#ifdef HAVE_GLOB_H
# include <glob.h>
#endif

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#include "common.h"

char *ldapconfigfile[] =
    { "/etc/openldap/ldap.conf", "/etc/ldap/ldap.conf", "~/.ldaprc",
	"./.ldaprc", NULL
};

char *userfilterfmt = "(&(objectClass=posixAccount)(uid=%s))";
char *userattrs[] = { "dn", "uid", "gecos", "loginShell", NULL };

int get_ldap_config(LDAPConfig * config)
{
	FILE *file;
	char **filename = ldapconfigfile;
	char *filenameptr;
	glob_t globbuf;
	char *buffer;
	char *ptr;
	size_t bufsize = 0;
	size_t count;
	int line;

	
	debug_msg(2, "get_ldap_config(%p)\n", config);

	config->scope = LDAP_SCOPE_SUBTREE;

	memset(&globbuf, 0, sizeof(glob_t));

	while (*filename) {
		filenameptr = *filename;
		if (**filename == '.' || **filename == '~') {
			if (glob(*filename, GLOB_TILDE, NULL, &globbuf) == 0)
				filenameptr = globbuf.gl_pathv[0];
		} else
			filenameptr = *filename;

		if (!(file = fopen(filenameptr, "r"))) {
			debug_msg(2, "fopen(%s, r): %s.\n", *filename,
				  strerror(errno));
			filename++;
			continue;
		}
		debug_msg(2, "fopen(%s, r)\n", *filename);

		if (globbuf.gl_pathv != NULL
		    && filenameptr == globbuf.gl_pathv[0])
			globfree(&globbuf);

		line = 0;
		while ((count = getline(&buffer, &bufsize, file)) != -1) {
			line++;
			if (*buffer == '#' || *buffer == '\n')
				continue;

			if ((ptr = strchr(buffer, '\n')))
				*ptr = '\0';

			ptr = buffer;
			while (*ptr != ' ' && *ptr != '\t' && *ptr != '\0')
				ptr++;

			*ptr++ = '\0';

			if (*ptr != '\0')
				while (*ptr == ' ' || *ptr == '\t'
				       || *ptr == '\0')
					ptr++;

			if (!strcasecmp("uri", buffer)) {
				if (*ptr)
					config->url = strdup(ptr);
				else
					config->url = strdup(ptr);
			} else if (!strcasecmp("base", buffer)) {
				if (*ptr)
					config->base = strdup(ptr);
				else
					config->base = NULL;
			} else if (!strcasecmp("tls_cacert", buffer)) {
				if (*ptr)
					config->tls_cacertfile = strdup(ptr);
				else
					config->tls_cacertfile = NULL;
			} else if (!strcasecmp("tls_cacertdir", buffer)) {
				if (*ptr)
					config->tls_cacertdir = strdup(ptr);
				else
					config->tls_cacertdir = NULL;
			} else if (!strcasecmp("tls_reqcert", buffer)) {
				if (*ptr)
					config->tls_reqcert = strdup(ptr);
				else
					config->tls_reqcert = NULL;
			}
		}
		fclose(file);
		filename++;
	}
	return (0);
}

int
rebindproc(LDAP * ld, LDAP_CONST char *url, ber_tag_t request, ber_int_t msgid,
	   void *params)
{
	int err;
	LDAPConfig *config = (LDAPConfig *) params;

	debug_msg(3,
		  "rebindproc(ld=%p, url=%s, request=%d, msgid=%d, params=%p)\n",
		  ld, url, request, msgid, params);

	debug_msg(3, "ldap_sasl_bind_s(dn:%s)\n", config->auth.userdn);
	err =
	    ldap_sasl_bind_s(ld, config->auth.userdn, LDAP_SASL_SIMPLE,
			     &config->auth.cred, NULL, NULL, NULL);
	if (err != LDAP_SUCCESS) {
		fprintf(stderr, "ldap_sasl_bind_s(dn:%s): %s\n",
			config->auth.userdn, ldap_err2string(err));
		return (err);
	}

	return (err);
}

LDAP *ldap_connect(LDAPConfig * config)
{
	LDAP *ld;
	int opt;
	int err;
	BerValue cred = {.bv_val = NULL,.bv_len = 0 };
	char *password;
	size_t password_bufsize = PASSWORD_BUFSIZE;

	USEREntry *authuser = NULL;

	password = (char *)malloc(password_bufsize);

	debug_msg(3, "ldap_connect(config=%p)\n", config);

	debug_msg(1, "ldap_initialize(%s)\n", config->url);
	if ((err = ldap_initialize(&ld, config->url)) != LDAP_SUCCESS) {
		fprintf(stderr, "ldap_initialize(%s): %s\n", config->url,
			ldap_err2string(err));
		return NULL;
	}

	/* Set the ldap version to 3 if it hasn't been set */
	if (config->version == 0)
		config->version = LDAP_VERSION3;

	debug_msg(3, "ldap_set_option(%p, LDAP_OPT_PROTOCOL_VERSION, %d)\n", ld,
		    config->version);
	if ((err =
	     ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION,
			     &config->version)) != LDAP_SUCCESS) {
		fprintf(stderr,
			"ldap_set_option(%p, LDAP_OPT_PROTOCOL_VERSION, %d): %s\n",
			ld, config->version, ldap_err2string(err));
	}

	/* XXX Not sure why, but LDAP_OPT_X_TLS_CACERTFILE only seems to work globally, not on a handle */
	debug_msg(3, "ldap_set_option(NULL, LDAP_OPT_X_TLS_CACERTFILE, %s)\n",
		    config->tls_cacertfile);
	if ((err =
	     ldap_set_option(NULL, LDAP_OPT_X_TLS_CACERTFILE,
			     config->tls_cacertfile)) != LDAP_SUCCESS) {
		fprintf(stderr,
			"ldap_set_option(NULL, LDAP_OPT_X_TLS_CACERT, %s): %s\n",
			config->tls_cacertfile, ldap_err2string(err));
	}

	/* XXX Not sure why, but LDAP_OPT_X_TLS_CACERTDIR only seems to work globally, not on a handle */
	debug_msg(3, "ldap_set_option(NULL, LDAP_OPT_X_TLS_CACERTDIR, %s)\n",
		    config->tls_cacertdir);
	if ((err =
	     ldap_set_option(NULL, LDAP_OPT_X_TLS_CACERTDIR,
			     config->tls_cacertdir)) != LDAP_SUCCESS) {
		fprintf(stderr,
			"ldap_set_option(NULL, LDAP_OPT_X_TLS_CACERTDIR, %s): %s\n",
			config->tls_cacertdir, ldap_err2string(err));
	}

	opt = LDAP_OPT_X_TLS_NEVER;
	if (strcasecmp("demand", config->tls_reqcert) == 0
	    || strcasecmp("hard", config->tls_reqcert) == 0)
		opt = LDAP_OPT_X_TLS_DEMAND;
	else if (!(strcasecmp("allow", config->tls_reqcert)))
		opt = LDAP_OPT_X_TLS_ALLOW;
	else if (!(strcasecmp("try", config->tls_reqcert)))
		opt = LDAP_OPT_X_TLS_TRY;

	/* XXX Not sure why, but LDAP_OPT_X_TLS_REQUIRE_CERT only seems to work globally, not on a handle */
	debug_msg
	    (3, "ldap_set_option(NULL, LDAP_OPT_X_TLS_REQUIRE_CERT, %d \"%s\"))\n",
	     opt, config->tls_reqcert);
	if ((err =
	     ldap_set_option(NULL, LDAP_OPT_X_TLS_REQUIRE_CERT,
			     &opt)) != LDAP_SUCCESS) {
		fprintf(stderr,
			"ldap_set_option(NULL, LDAP_OPT_X_TLS_REQUIRE_CERT, %d): %s\n",
			opt, ldap_err2string(err));
	}

	if (opt != LDAP_OPT_X_TLS_NEVER) {
		debug_msg(1, "ldap_start_tls_s(%p, NULL, NULL)\n", ld);
		err = ldap_start_tls_s(ld, NULL, NULL);
		switch (err) {
			char *errmsg;
		case LDAP_SUCCESS:
			break;
		case LDAP_CONNECT_ERROR:
			ldap_get_option(ld, LDAP_OPT_DIAGNOSTIC_MESSAGE,
					(void *)&errmsg);
			fprintf(stderr,
				"ldap_start_tls_s(%p, NULL, NULL): %s\n", ld,
				errmsg);
			ldap_memfree(errmsg);
			return NULL;
		default:
			fprintf(stderr,
				"ldap_start_tls_s(%p, NULL, NULL): %s\n", ld,
				ldap_err2string(err));
			return NULL;
		}
	}

	debug_msg(2, "ldap_sasl_bind_s(anonymous)\n");
	err =
	    ldap_sasl_bind_s(ld, NULL, LDAP_SASL_SIMPLE, &cred, NULL, NULL,
			     NULL);
	if (err != LDAP_SUCCESS) {
		fprintf(stderr, "ldap_sasl_bind_s(anonymous): %s\n",
			ldap_err2string(err));
		return (NULL);
	}

	if ((authuser =
	     get_ldap_user_entry(ld, config, config->auth.user)) == NULL) {
		printf("Unable to find current user \"%s\" in LDAP\n",
		       config->auth.user);
		ldap_unbind_ext_s(ld, NULL, NULL);
		return NULL;
	}

	config->auth.userdn = authuser->dn;

	if ((my_getpass
	     ("Enter LDAP Password: ", &password, &password_bufsize,
	      stdin) == -1))
		return NULL;

	if (strlen(password) > 0 && password[strlen(password) - 1] == '\n')
		password[strlen(password) - 1] = '\0';

	config->auth.cred.bv_val = strdup(password);
	config->auth.cred.bv_len = strlen(config->auth.cred.bv_val);

	free(password);

#ifdef DEBUG_AUTH
	debug_msg(3, "ldap_sasl_bind_s(dn:%s, pw:%s)\n",
		  config->auth.userdn, config->auth.cred.bv_val);
#else
	debug_msg(3, "ldap_sasl_bind_s(dn:%s)\n", config->auth.userdn);
#endif
	err =
	    ldap_sasl_bind_s(ld, config->auth.userdn, LDAP_SASL_SIMPLE,
			     &config->auth.cred, NULL, NULL, NULL);
	if (err != LDAP_SUCCESS) {
		fprintf(stderr, "ldap_sasl_bind_s(dn:%s): %s\n",
			config->auth.userdn, ldap_err2string(err));
		return (NULL);
	}

	ldap_set_rebind_proc(ld, rebindproc, (void *)config);

	return ld;
}

void free_ldap_user_entry(USEREntry * user)
{
	if (user->dn) {
		free(user->dn);
		user->dn = NULL;
	}
	if (user->uid) {
		free(user->uid);
		user->uid = NULL;
	}
	if (user->gecos) {
		free(user->gecos);
		user->uid = NULL;
	}
	if (user->loginShell) {
		free(user->loginShell);
		user->loginShell = NULL;
	}
	if (user) {
		free(user);
		user = NULL;
	}
}

USEREntry *get_ldap_user_entry(LDAP * ld, LDAPConfig * config, char *uid)
{
	USEREntry *user;
	LDAPMessage *result;
	LDAPMessage *entry;
	char filter[PASSWORD_BUFSIZE];
	int count;
	char *dn;
	BerValue **vals;

	debug_msg(2, "get_ldap_user_entry()\n");

	user = malloc(sizeof(USEREntry));

	sprintf(filter, userfilterfmt, uid);

	debug_msg(3, "ldap_search_ext_s(\n\t\tfilter->%s\n"
		  "\t\tscope->%d\n\t\tbase->%s\n\t)\n", filter,
		  config->scope, config->base);

	ldap_search_ext_s(ld,		// LDAP         *ld
			  config->base,	// char         *base
			  config->scope,// int           scope
			  filter,	// char         *filter
			  userattrs,	// char         *attrs[]
			  0,		// int           attrsonly
			  NULL,		// LDAPControl **serverctrls
			  NULL,		// LDAPControl **clientctrls
			  NULL,		// struct timeval *timeout
			  0,		// int           sizelimit
			  &result	// LDAPMessage  **result
	    );

	count = ldap_count_entries(ld, result);

	debug_msg(3, "ldap_count_entries() == %d\n", count);

	if (count == 0) {
		printf("Unable to find user \"%s\".\n", uid);
		ldap_msgfree(result);
		return (NULL);
	} else if (count > 1) {
		printf("Found multiple entries for user \"%s\".\n", uid);
		ldap_msgfree(result);
		return (NULL);
	}

	entry = ldap_first_entry(ld, result);

	dn = ldap_get_dn(ld, entry);
	user->dn = strdup(dn);
	ldap_memfree(dn);

	vals = ldap_get_values_len(ld, entry, "uid");
	if (ldap_count_values_len(vals) > 0) {
		user->uid = strdup(vals[0]->bv_val);
	}
	ldap_value_free_len(vals);

	vals = ldap_get_values_len(ld, entry, "gecos");
	if (ldap_count_values_len(vals) > 0) {
		user->gecos = strdup(vals[0]->bv_val);
	}
	ldap_value_free_len(vals);

	vals = ldap_get_values_len(ld, entry, "loginShell");
	if (ldap_count_values_len(vals) > 0) {
		user->loginShell = strdup(vals[0]->bv_val);
	}
	ldap_value_free_len(vals);

	ldap_memfree(entry);

	while (((entry = ldap_next_entry(ld, entry)) != NULL))
		ldap_memfree(entry);

	return (user);
}
